#include "pqtree.h"
#include "dataset.h"
#include <getopt.h>
#include <stdlib.h>

uintvector
find_ones(const boolvector & v)
{
	uint32_t cnt = 0;
	for (uint32_t i = 0; i < v.size(); i++)
		cnt += v[i];

	uintvector res(cnt);

	for (uint32_t i = 0, j = 0; i < v.size(); i++) {
		//printf("%d ", int(v[i]));
		if (v[i]) {
			res[j] = i;
			j++;
		}
	}

	//printf("\n");

	return res;
}

uintvector
find_zeros(const boolvector & v)
{
	uint32_t cnt = 0;
	for (uint32_t i = 0; i < v.size(); i++)
		cnt += 1 - v[i];

	uintvector res(cnt);

	for (uint32_t i = 0, j = 0; i < v.size(); i++) {
		if (!v[i]) {
			res[j] = i;
			j++;
		}
	}

	return res;
}

struct block {
	uintvector row;
	uintvector col;
};

struct tile
{
	double gain;
	block raw;
	block norm;
};


class tiler
{
	public:
		tiler(dataset & d, dataset & seed, bool cyclic) :
			m_data(d), 
			m_cover(d.rowcnt(), d.colcnt()),
			m_coltree(d.colcnt()),
			m_rowtree(d.rowcnt()),
			m_seed(seed),
			m_cyclic(cyclic)
			{}

		void iterate(const uintvector & seed, tile & t);
		void iterate_cyclic(const uintvector & seed, tile & t);

		// parallel-safe versions
		void iterate_safe(const uintvector & seed, tile & t) const;
		void iterate_cyclic_safe(const uintvector & seed, tile & t) const;

		void step_exhaustive(tile & t);
		void find_cover(uint32_t k);

		void save_tiles(const char *name);
		void save_order(const char *name);

	protected:
		double find_optimal_col(pqtree & tree, const uintvector & r, boolvector & c) const;
		double find_optimal_row(pqtree & tree, const uintvector & r, boolvector & c) const;
		double find_optimal_reverse_col(pqtree & tree, const uintvector & r, boolvector & c) const;
		double find_optimal_reverse_row(pqtree & tree, const uintvector & r, boolvector & c) const;

		dataset & m_data;
		dataset m_cover;
		pqtree m_coltree;
		pqtree m_rowtree;

		dataset & m_seed;

		bool m_cyclic;

		typedef std::list<block> blocklist;

		blocklist m_blocks;
};


void
tiler::save_tiles(const char *name)
{
	FILE *f = fopen(name, "w");

	for (blocklist::iterator it = m_blocks.begin(); it != m_blocks.end(); ++it) {
		for (uint32_t i = 0; i < it->row.size(); i++)
			fprintf(f, "%d ", it->row[i]);
		fprintf(f, "\n");
		for (uint32_t i = 0; i < it->col.size(); i++)
			fprintf(f, "%d ", it->col[i]);
		fprintf(f, "\n");
	}

	fclose(f);
}

void
tiler::save_order(const char *name)
{
	FILE *f = fopen(name, "w");
	m_rowtree.print_order(f);
	fprintf(f, "\n");
	m_coltree.print_order(f);
	fprintf(f, "\n");
	fclose(f);
}




double
tiler::find_optimal_col(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), -double(r.size()));
	m_data.sum_rows(r, 2, w);
	m_cover.sum_rows(r, 1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	return gain;
}

double
tiler::find_optimal_row(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), -double(r.size()));
	m_data.sum_cols(r, 2, w);
	m_cover.sum_cols(r, 1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	return gain;
}

double
tiler::find_optimal_reverse_col(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), double(r.size()));
	m_data.sum_rows(r, -2, w);
	m_cover.sum_rows(r, -1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	double tw = 0;
	for (uint32_t i = 0; i < w.size(); i++)
		tw += w[i];

	return gain - tw;
}

double
tiler::find_optimal_reverse_row(pqtree & tree, const uintvector & r, boolvector & c) const
{
	doublevector w(m_data.colcnt(), double(r.size()));
	m_data.sum_cols(r, -2, w);
	m_cover.sum_cols(r, -1, w);

	tree.set_weights(w);
	double gain = tree.find_max();
	tree.assign_optimal(c);

	double tw = 0;
	for (uint32_t i = 0; i < w.size(); i++)
		tw += w[i];

	return gain - tw;
}

void
tiler::iterate_safe(const uintvector & seed, tile & t) const
{
	pqtree rt = pqtree(m_rowtree);
	pqtree ct = pqtree(m_coltree);

	double last_gain = -1;
	t.gain = 0;

	t.raw.col = seed;

	while (t.gain > last_gain) {
		last_gain = t.gain;

		boolvector br(m_data.rowcnt());
		find_optimal_row(rt, t.raw.col, br);
		t.raw.row = find_ones(br);


		boolvector bc(m_data.colcnt());
		t.gain = find_optimal_col(ct, t.raw.row, bc);
		t.raw.col = find_ones(bc);

	}

	t.norm = t.raw;
}



void
tiler::iterate(const uintvector & seed, tile & t)
{
	double last_gain = -1;
	t.gain = 0;

	t.raw.col = seed;

	while (t.gain > last_gain) {
		last_gain = t.gain;

		boolvector br(m_data.rowcnt());
		find_optimal_row(m_rowtree, t.raw.col, br);
		t.raw.row = find_ones(br);


		boolvector bc(m_data.colcnt());
		t.gain = find_optimal_col(m_coltree, t.raw.row, bc);
		t.raw.col = find_ones(bc);

	}

	t.norm = t.raw;
}

void
tiler::iterate_cyclic_safe(const uintvector & seed, tile & t) const
{
	pqtree rt = pqtree(m_rowtree);
	pqtree ct = pqtree(m_coltree);
	double last_gain = -1;
	t.gain = 0;

	t.raw.col = seed;
	t.norm.col = seed;

	double g1, g2;
	while (t.gain > last_gain) {
		last_gain = t.gain;

		boolvector br1(m_data.rowcnt());
		boolvector br2(m_data.rowcnt());
		g1 = find_optimal_row(rt, t.norm.col, br1);
		g2 = find_optimal_reverse_row(rt, t.norm.col, br2);
		//printf("row: %f %f\n", g1, g2);

		if (g1 > g2) {
			t.norm.row = t.raw.row = find_ones(br1);
		}
		else {
			t.raw.row = find_ones(br2);
			t.norm.row = find_zeros(br2);
		}

		//for (uint32_t i = 0; i < t.norm.row.size(); i++)
			//printf("%d ", t.norm.row[i]);
		//printf("\n");


		boolvector bc1(m_data.colcnt());
		boolvector bc2(m_data.colcnt());
		g1 = find_optimal_col(ct, t.norm.row, bc1);
		g2 = find_optimal_reverse_col(ct, t.norm.row, bc2);
		//printf("col: %f %f\n", g1, g2);
		if (g1 > g2) {
			t.norm.col = t.raw.col = find_ones(bc1);
		}
		else {
			t.raw.col = find_ones(bc2);
			t.norm.col = find_zeros(bc2);
		}

		t.gain = std::max(g1, g2);
	}
}



void
tiler::iterate_cyclic(const uintvector & seed, tile & t)
{
	double last_gain = -1;
	t.gain = 0;

	t.raw.col = seed;
	t.norm.col = seed;

	double g1, g2;
	while (t.gain > last_gain) {
		last_gain = t.gain;

		boolvector br1(m_data.rowcnt());
		boolvector br2(m_data.rowcnt());
		g1 = find_optimal_row(m_rowtree, t.norm.col, br1);
		g2 = find_optimal_reverse_row(m_rowtree, t.norm.col, br2);
		//printf("row: %f %f\n", g1, g2);

		if (g1 > g2) {
			t.norm.row = t.raw.row = find_ones(br1);
		}
		else {
			t.raw.row = find_ones(br2);
			t.norm.row = find_zeros(br2);
		}

		//for (uint32_t i = 0; i < t.norm.row.size(); i++)
			//printf("%d ", t.norm.row[i]);
		//printf("\n");


		boolvector bc1(m_data.colcnt());
		boolvector bc2(m_data.colcnt());
		g1 = find_optimal_col(m_coltree, t.norm.row, bc1);
		g2 = find_optimal_reverse_col(m_coltree, t.norm.row, bc2);
		//printf("col: %f %f\n", g1, g2);
		if (g1 > g2) {
			t.norm.col = t.raw.col = find_ones(bc1);
		}
		else {
			t.raw.col = find_ones(bc2);
			t.norm.col = find_zeros(bc2);
		}

		t.gain = std::max(g1, g2);
	}
}


void
tiler::step_exhaustive(tile & best)
{
  best.gain = -1;
#pragma omp parallel
  {
    tile priv_best;
    priv_best.gain = -1;
#pragma omp for nowait
    for (uint32_t i = 0; i < m_seed.rowcnt(); i++) {
      const uintset & rs = m_seed.row(i);
      uintvector seed(rs.size());
      std::copy(rs.begin(), rs.end(), seed.begin());
	    
      //printf("SEED: %d\n", i);
      tile t;
      if (m_cyclic)
	iterate_cyclic_safe(seed, t);
      else
	iterate_safe(seed, t);
      
      if (t.gain >= priv_best.gain)
	priv_best = t;
    }
#pragma omp critical
    {
      if (priv_best.gain >= best.gain) best = priv_best;
    }
  }
}

void
tiler::find_cover(uint32_t k)
{
	for (uint32_t i = 0; i < k; i++) {
		tile t;
		step_exhaustive(t);

		m_rowtree.reduce(t.raw.row);
		m_coltree.reduce(t.raw.col);
		//m_coltree.print();

		m_data.remove_tile(t.norm.row, t.norm.col);
		m_cover.add_tile(t.norm.row, t.norm.col);

		m_blocks.push_back(t.norm);
	}
}



int
main(int argc, char **argv)
{
	static struct option longopts[] = {
		{"out",             required_argument,  NULL, 'o'},
		{"in",              required_argument,  NULL, 'i'},
		{"seed",            required_argument,  NULL, 's'},
		{"order",           required_argument,  NULL, 'r'},
		{"comps",           required_argument,  NULL, 'k'},
		{"cyclic",           no_argument,       NULL, 'c'},
		{ NULL,             0,                  NULL,  0 }
	};

	char *inname = NULL;
	char *outname = NULL;
	char *ordername = NULL;
	bool cyclic = false;
	uint32_t groupcnt = 1;

	dataset *seed = 0;


	int ch;
	while ((ch = getopt_long(argc, argv, "chk:o:i:s:r:", longopts, NULL)) != -1) {
		switch (ch) {
			case 'h':
				printf("Usage: %s -i <input file> [-o <output file>] -k <number> [-h]\n", argv[0]);
				printf("  -h    print this help\n");
				printf("  -i    input file\n");
				printf("  -o    output file\n");
				printf("  -c    allow cyclic order\n");
				printf("  -k    number of groups\n");
				return 0;
				break;
			case 'i':
				inname = optarg;
				break;
			case 'c':
				cyclic = true;
				break;
			case 'o':
				outname = optarg;
				break;
			case 'r':
				ordername = optarg;
				break;
			case 's':
				seed = new dataset(optarg);
				break;
			case 'k':
				groupcnt = atoi(optarg);
				break;
		}

	}


	/*
	pqtree pq(6);
	pq.print();
	uintvector r(3);
	r[0] = 3;
	r[1] = 4;
	r[2] = 5;
	pq.reduce(r);
	pq.print();

	return 0;
	*/

	if (inname == NULL) {
		printf("Missing input file\n");
		return 1;
	}


	dataset data(inname);	

	if (seed == 0) {
		seed = new dataset(data.colcnt(), data.colcnt());
		seed->add_diagonal();
	}

	tiler t(data, *seed, cyclic);

	t.find_cover(groupcnt);

	if (outname)
		t.save_tiles(outname);
    if (ordername)
		t.save_order(ordername);

	return 0;
}
