#ifndef PQTREE_H
#define PQTREE_H

#include "defines.h"
#include <stdio.h>

enum pqtype {pqtype_p, pqtype_q, pqtype_leaf};
enum pqlabel {pqlabel_empty, pqlabel_partial, pqlabel_full};

struct pqnode {
	pqnode():
		parent(0), child_count(0), first(0), last(0), left(0), right(0),
		queue(0), qid(0),
		weight(0), side_weight(0), inner_weight(0),
		side_first(0), side_next(0), inner_first(0), inner_next(0),
		next_free(0)
	{
	}

	uint32_t id;

	pqnode *parent;
	pqtype type;
	pqlabel label;

	uint32_t child_count;
	pqnode *first;
	pqnode *last;
	pqnode *left;
	pqnode *right;

	pqnode *queue;
	uint32_t qid; // used for determining, if in queue
	uint32_t pertinent_leaf_count;
	uint32_t visited_count;

	double weight;
	double side_weight;
	double inner_weight;


	pqnode *side_first;
	pqnode *side_next;
	pqnode *inner_first;
	pqnode *inner_next;

	pqnode *next_free;

	void
	print(uint32_t indent)
	{
		for (uint32_t i = 0; i < indent; i++)
			printf(" ");
		if (type == pqtype_leaf)
			printf("%d %.0f %.0f %.0f %d\n", id, weight, side_weight, inner_weight, label);
		else {
			if (type == pqtype_p)
				printf("p %.0f %.0f %.0f\n", weight, side_weight, inner_weight);
			else
				printf("q %.0f %.0f %.0f\n", weight, side_weight, inner_weight);
			for (pqnode *c = first; c; c = c->right) 
				c->print(indent + 1);
		}
	}

	void
	print_order(FILE *f)
	{
		if (type == pqtype_leaf)
			fprintf(f, "%d ", id);
		else {
			for (pqnode *c = first; c; c = c->right) 
				c->print_order(f);
		}
	}

	void
	detach() // detach this from a parent.
	{
		if (parent == 0) return;
		if (left)
			left->right = right;
		if (right)
			right->left = left;
		parent->child_count--;
		if (parent->first == this)
			parent->first = right;
		if (parent->last == this)
			parent->last = left;

	}

    //insert this to as a child of par after aft (aft can be None)
	void
	insert_after(pqnode *par, pqnode *aft)
	{
		parent = par;
		par->child_count++;
		if (aft == 0) {
			left = 0;
			right = par->first;
			par->first = this;
		}
		else {
			right = aft->right;
			aft->right = this;
			left = aft;
		}
		if (right)
			right->left = this;
		else
			par->last = this;
	}

    // moves empty nodes from s to target
    // maintains order, adds to the left
	void
	migrate_empty(pqnode *target)
	{
		pqnode *c = last;
		while (c) {
			pqnode *cnext = c->left;
			if (c->label == pqlabel_empty) {
				//printf("me: %d\n", c->id);
				c->detach();
				c->insert_after(target, 0);
			}
			c = cnext;
		}
	}

    // moves full nodes from s to target
    // maintains order, adds to the right
	void
	migrate_full(pqnode *target)
	{
		pqnode *c = first;
		while (c) {
			pqnode *cnext = c->right;
			if (c->label == pqlabel_full) {
				//printf("mf: %d\n", c->id);
				c->detach();
				c->insert_after(target, target->last);
			}
			c = cnext;
		}
	}
    // moves all nodes from s to target
    // reverses order
	void
	migrate_mirror(pqnode *target)
	{
		while (last) {
			pqnode *c = last;
			c->detach();
			c->insert_after(target, target->last);
		}
	}

    // moves all nodes from s to target after aft
	void
	migrate_after(pqnode *target, pqnode *aft)
	{
		while (first) {
			pqnode *c = first;
			c->detach();
			c->insert_after(target, aft);
			aft = c;
		}
	}

	// reverses the order of children
	void
	reverse()
	{
		pqnode *stop = last;
		while (first != stop) {
			pqnode *c = first;
			c->detach();
			c->insert_after(this, stop);
		}
	}

	void assign_interval_side(boolvector & out);
	void assign_interval_full(boolvector & out);
	void assign_interval_inner(boolvector & out);

	void compute_total_weight();
	void compute_side_weight_perm();
	void compute_best_inner_child();
	void compute_inner_weight_perm();
	void compute_side_weight_queue();
	void compute_inner_weight_queue();

};

typedef std::vector<pqnode> pqvector;

struct queue {
	pqnode *first;
	pqnode *last;
	uint32_t qid;

	void reset() {qid++; first = 0; last = 0;}

	bool empty() {return first == 0;}

	pqnode *
	pop()
	{
		pqnode *n = first;
		if (n) {first = n->queue;}
		return n;
	}


	void
	push(pqnode *n)
	{
		if (n == 0 || n->qid == qid) return;
		n->qid = qid;
		n->queue = 0;

		if (first == 0) {
			first = last = n;
		}
		else {
			last->queue = n;
			last = n;
		}
	}
};


class
pqtree {
	public:
		pqtree(uint32_t cnt);
		pqtree(const pqtree & t);

		bool reduce(uintvector & order);
		double find_max();
		void assign_optimal(boolvector & out);
		void set_weights(doublevector & weights);

		void print() {m_sentinel->first->print(0);}
		void print_order(FILE *f) {m_sentinel->first->print_order(f);}



	protected:
		void reset_counter();
		pqnode * find_root(uintvector & order);
		bool process(pqnode *n, bool isroot);

		void template_p2(pqnode *n);
		void template_p4(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *p1);
		void template_p6(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *p1, pqnode *p2);
		void template_p3(pqnode *n, uint32_t empty_count, uint32_t full_count);
		void template_p5(pqnode *n, uint32_t empty_count, uint32_t full_count, pqnode *p1);

		void template_q1(pqnode *n);
		void template_q2(pqnode *n, pqnode *p1);
		void template_q3(pqnode *n, pqnode *p1, pqnode *p2);

		pqnode *new_node(pqtype t);
		void release_node(pqnode *n);
		
		pqnode *m_sentinel;
		queue m_queue;
		pqvector m_leaves;
		pqvector m_nodes;
		pqnode *m_freenode;
};


#endif
